// naumann@stce.rwth-aachen.de

#include "linear_system.hpp"
#include "linear_solver_qr.hpp"

#include<cassert>
#include<iostream>
#include<cmath>
#include<limits>

Linear_System::VT x(const Linear_System::MT& A, const Linear_System::VT& b) {
  auto n=b.size();
  Linear_System lsys(n);
  lsys.A()=A; lsys.b()=b; 
  Linear_Solver_QR lsol;
  lsol.solve(lsys);
  return lsys.x();
}

Linear_System::MT dxdb(const Linear_System::MT& A, Linear_System::MT b) {
  auto n=b.size();
  Linear_System::MT r(n,n);
  for (auto i=0;i<n;i++) {
    const Linear_System::T h=fabs(b(i))<=1
      ? sqrt(std::numeric_limits<Linear_System::T>::epsilon())
      : sqrt(std::numeric_limits<Linear_System::T>::epsilon())*fabs(b(i));
    b(i)-=h; 
    Linear_System::VT xm=x(A,b); 
    b(i)+=2*h; 
    Linear_System::VT xp=x(A,b); 
    b(i)-=h;
    for (auto j=0;j<n;j++) r(j,i)=(xp(j)-xm(j))/(2*h);
  }
  return r;
}

int main(int argc, char* argv[]) {
  assert(argc==2); int n=std::stoi(argv[1]); 
  Linear_System::MT A=Linear_System::MT::Random(n,n); 
  Linear_System::VT b=Linear_System::VT::Random(n); 
  std::cout << "A=" << std::endl << A << std::endl;
  // should be equal to inverse
  std::cout << "dxdb=" << std::endl << dxdb(A,b) << std::endl << "=A^-1=" 
            << std::endl << A.partialPivLu().inverse() << std::endl;
  // should be equal to identity
  std::cout << "A*dxdb=" << std::endl << A*dxdb(A,b) << std::endl;
  // should be equal to condition
  std::cout << "||A||*||dxdb||=" << A.norm()*dxdb(A,b).norm() << "=cond(A)="
            << A.norm()*A.partialPivLu().inverse().norm() << std::endl;
  return 0;
}

